from scapy.all import * # Scapy
import threading
import signal
import sys

import dash # Dash framework
import dash_html_components as html
import dash_core_components as dcc
from dash.dependencies import Input,Output,State
import dash_table

import pandas as pd # Pandas

import plotly.graph_objects as go # Plotly

# inicializacia Pandas DataFrame,index = id
pkt_data = pd.DataFrame(columns = ["id","pktnum","srcIP","srcSERVICE","dstIP","dstSERVICE","hasIP","hasTCP"])
pkt_data.set_index("id",inplace = True)
Ethernet_data = pd.DataFrame(columns = ["id","srcMAC","dstMAC","type"])
Ethernet_data.set_index("id",inplace = True)
IP_data = pd.DataFrame(columns = ["id","version","ttl","protocol","srcIP","dstIP"])
IP_data.set_index("id",inplace = True)
TCP_data = pd.DataFrame(columns = ["id","srcPORT","dstPORT","seq","ack"])
TCP_data.set_index("id",inplace = True)

Ethernet_frame_types = {
    0x0800 : "IPv4",
    0x0806 : "ARP",
    0x0842 : "Wake on LAN",
    0x8035 : "RARP",
    0x809b : "AppleTalk (Ethertalk)",
    0x80f3 : "AARP",
    0x8137 : "IPX",
    0x86dd : "IPv6",
    0x8808 : "Ethernet flow control",
    0x8847 : "MPLS unicast",
    0x8848 : "MPLS multicast",
    0x8863 : "PPPoE Discovery Stage",
    0x8864 : "PPPoE Session Stage",
    0x888e : "EAP over LAN (IEEE 802.1X)",
    0x88a4 : "EtherCAT",
    0x88cc : "LLDP",
    0x8906 : "FCoE",
    0x8914 : "FCoE Initialization Protocol"
}

IPprotocols = {
    1   : "ICMP Internet Control Message Protocol",
    2   : "IGMP Internet Group Management Protocol",
    3   : "GGP Gateway-to-Gateway Protocol",
    4   : "IP in IP (encapsulation)",
    5   : "ISP Internet Stream Protocol",
    6   : "TCP Transmission Control Protocol",
    8   : "EGP Exterior Gateway Protocol",
    9   : "IGP Interior Gateway Protocol",
    17  : "UDP User Datagram Protocol",
    20  : "HMP Host Monitoring Protocol",
    27  : "RDP Reliable Data Protocol",
    47  : "GRE Generic Routing Encapsulation",
    50  : "ESP Encapsulating Security Payload",
    51  : "AH Authentication Header",
    66  : "MIT Remote Virtual Disk Protocol",
    88  : "EIGRP Enhanced Interior Gateway Routing Protocol",
    89  : "OSPF Open Shortest Path First",
    133 : "FC Fibre Channel"
}

PORTSRV = {
    7    : "ECHO protokol",
    13   : "DAYTIME protokol",
    20   : "FTP prenos dát",
    21   : "FTP správa",
    22   : "SSH protokol",
    23   : "Telnet protokol",
    25   : "SMTP protokol",
    37   : "TIME protokol",
    42   : "WINS služba",
    43   : "WHOIS protokol",
    52   : "XNS Time protokol",
    53   : "DNS služba",
    56   : "RAP protokol",
    67   : "DHCP protokol",
    68   : "DHCP protokol",
    69   : "TFTP protokol",
    79   : "FINGER protokol",
    80   : "HTTP protokol",
    88   : "Kerberos autentifikácia",
    107  : "Vzdialená Telnet služba",
    109  : "POP2 protokol",
    110  : "POP3 protokol",
    115  : "SFTP protokol",
    118  : "SQL služby",
    123  : "NTP protokol",
    137  : "NetBIOS Name služba",
    138  : "NetBIOS Datagram služba",
    139  : "NetBIOS Session služba",
    143  : "IMAP protokol",
    152  : "BFTP Background File Transfer Program",
    156  : "SQL služba",
    161  : "SNMP protokol",
    162  : "SNMP protokol",
    170  : "Network PostScript",
    194  : "IRC služba",
    220  : "IMAPv3",
    280  : "HTTP manažment",
    389  : "LDAP protokol",
    443  : "HTTPS protokol",
    464  : "Kerberos nastavenie/zmena hesla",
    465  : "SMTP cez TLS/SSL",
    500  : "ISAKMP/IKE",
    520  : "RIP protokol",
    530  : "RPC protokol",
    543  : "klogin Kerberos login",
    544  : "kshell Kerberos shell",
    546  : "DHCPv6 klient",
    547  : "DHCPv6 server",
    554  : "RTSP protokol",
    631  : "CUPS služba",
    636  : "LDAP cez TLS/SSL",
    901  : "Samba WAT",
    903  : "VMware konzola",
    989  : "FTP data cez TLS/SSL",
    990  : "FTP kontrola cez TLS/SSL",
    995  : "POP3 cez TLS/SSL",
    3306 : "MySQL databáza",
    5432 : "PostgreSQL databáza"
}

pkt_num = 0 # cislo prijateho paketu
sel_pkt = 0 # cislo vybraneho paketu

def get_service(pkt,s = True):
    if s: # sluzba na zdroji
        try:
            if pkt.sport in PORTSRV.keys():
                svc = PORTSRV[pkt.sport]
            else:
                svc = str(pkt.sport)
        except AttributeError:
            svc = "Neznáma služba"
    else: # sluzba na cieli
        try:
            if pkt.dport in PORTSRV.keys():
                svc = PORTSRV[pkt.dport]
            else:
                svc = str(pkt.dport)
        except AttributeError:
            svc = "Neznáma služba"

    return svc

def pkt_recv(pkt):
    global pkt_num
    pkt_num += 1 # cislo prijateho paketu

    if pkt_num > 500: # ak bolo prijatych viac ako 500 paketov, DataFramy sa vymazu
        pkt_data.drop(pkt_data.index,inplace = True)
        Ethernet_data.drop(pkt_data.index,inplace = True)
        IP_data.drop(pkt_data.index,inplace = True)
        TCP_data.drop(pkt_data.index,inplace = True)
        pkt_num = 1

    #pkt.show() # zobrazenie obsahu paketu

    p = pkt[0][1]
    src_service_name = get_service(p) # sluzba na zdroji
    dst_service_name = get_service(p,False) # sluzba na cieli

    #if dst_service_name and dst_service_name in ["IMAP protokol","POP3 protokol","SMTP protokol"]:
    #    if pkt[TCP].payload:
    #        mail_pkt = str(pkt[TCP].payload)
    #        if "user" in mail_pkt.lower() or "pass" in mail_pkt.lower():
    #            print("Server: %s") % pkt[IP].dst
    #            print("%s") % pkt[TCP].payload
    #elif ARP in pkt:
    #    if pkt[ARP].op == 1: # who-has (request)
    #        print("Request: " + pkt[ARP].psrc + " sa pýta na " + pkt[ARP].pdst)
    #    if pkt[ARP].op == 2: # is-at (response)
    #        print("Response: " + pkt[ARP].hwsrc + " má adresu " + pkt[ARP].psrc)

    id = pkt_num # id = cislo prijateho paketu

    # typ Ethernet ramca
    try:
        frame_type = Ethernet_frame_types[pkt.type]
    except:
        frame_type = "Neznámy rámec"

    # Ethernet hlavička
    Ethernet_data.loc[id] = (pkt.src,pkt.dst,frame_type)
    # IP hlavička
    hasIP = False
    if pkt.haslayer(IP):
        hasIP = True
        try:
            protocol_name = IPprotocols[p.proto]
        except:
            protocol_name = "Neznámy protokol"
        IP_data.loc[id] = (p.version,p.ttl,protocol_name,p.src,p.dst)
    # TCP hlavička
    hasTCP = False
    if pkt.haslayer(TCP):
        hasTCP = True
        p = pkt[TCP]
        TCP_data.loc[id] = (p.sport,p.dport,p.seq,p.ack)
    
    # obsah paketu pre hlavnu tabulku
    p = pkt[0][1]
    pkt_data.loc[id] = (pkt_num,p.src,src_service_name,p.dst,dst_service_name,hasIP,hasTCP)

mIP = get_if_addr(conf.iface) # moja IP adresa

# inicializacia Dash aplikacie
app = dash.Dash(__name__)

# definicia Dash layoutu
app.layout = html.Div(
    # 2 riadky
    children = [html.Div(className = 'row', # 1. riadok ----------------------------------------------------------------------
                         children = [# 1. stlpec
                                     html.Div(children = [# 1. riadok
                                                          html.H1('Sliedič v.1'),
                                                          html.Div(className = 'block_container',
                                                                   style = {'margin' : 20},
                                                                   children = ['Moja IP adresa je:',html.Div(className = 'IPs',children = [mIP])]
                                                          )
                                              ]
                                     )
                         ]
                ),
                html.Div(className = 'row', # 2. riadok ----------------------------------------------------------------------
                         children = [# 1. stlpec
                                     html.Div(className = 'five columns',style = {'margin-left' : 20},
                                              children = [# 1. riadok
                                                          html.Div(children = [html.Button(id = 'obnov',children = ['Obnov'],
                                                                                                                   style = {'margin-bottom' : '10px',
                                                                                                                            'background-color' : '#cb1828',
                                                                                                                            'color' : 'white'}
                                                                                          )
                                                                              ]
                                                          ),
                                                          html.Div(children = [dash_table.DataTable(
                                                                                  id = 'table1',
                                                                                  fixed_rows = {'headers' : True,'data' : 0},
                                                                                  columns = [{'id' : 'pktnum','name' : 'Číslo paketu'},
                                                                                             {'id' : 'srcIP','name' : 'IP adresa zdroja'},
                                                                                             {'id' : 'srcSERVICE','name' : 'Služba na zdroji'},
                                                                                             {'id' : 'dstIP','name' : 'IP adresa cieľa'},
                                                                                             {'id' : 'dstSERVICE','name' : 'Služba na cieli'}],
                                                                                             #{'id' : 'hasIP','name' : 'hasIP'},
                                                                                             #{'id' : 'hasTCP','name' : 'hasTCP'}],
                                                                                  style_table = {'border' : 'grey solid',
                                                                                                 'width' : 1000,'height' : 400},
                                                                                  style_header = {'color' : 'lightgrey',
                                                                                                  'backgroundColor' : 'rgb(30,30,30)',
                                                                                                  'fontWeight' : 'bold'},
                                                                                  style_cell = {'color' : 'white',
                                                                                                'backgroundColor' : 'rgb(50,50,50)',
                                                                                                'border' : '0px solid',
                                                                                                'font-family' : 'Open Sans Semi',
                                                                                                'font-size' : '0.8em',
                                                                                                'textAlign' : 'center',
                                                                                                'min-width' : '70px'},
                                                                                  style_as_list_view = True,
                                                                                  style_data_conditional = [{'if' : {'row_index' : 'odd'},
                                                                                                             'backgroundColor' : 'rgb(70,70,70)'}],
                                                                                  page_action = 'none')
                                                                              ]
                                                          ),
                                                          html.Div(id='selected_packet_text',style = {'margin-top' : '10px'}),
                                                          html.Div(children = [html.Button(id = 'detaily',children = ['Detaily'],
                                                                                                                     style = {'margin-top' : '5px',
                                                                                                                              'margin-bottom' : '10px',
                                                                                                                              'background-color' : '#188818',
                                                                                                                              'color' : 'white'}
                                                                                          )
                                                                              ]
                                                          ),
                                                          html.Div('Ethernet hlavička'),
                                                          html.Div(style = {'margin-bottom' : '10px'},
                                                                   children = [dash_table.DataTable(
                                                                                  id = 'Ethernet_table',
                                                                                  fixed_rows = {'headers' : True,'data' : 0},
                                                                                  columns = [{'id' : 'srcMAC','name' : 'MAC adresa zdroja'},
                                                                                             {'id' : 'dstMAC','name' : 'MAC adresa cieľa'},
                                                                                             {'id' : 'type','name' : 'Typ rámca'}],
                                                                                  style_table = {'border' : 'grey solid',
                                                                                                 'width' : 500,'height' : 60},
                                                                                  style_header = {'color' : 'lightgrey',
                                                                                                  'backgroundColor' : 'rgb(30,30,30)',
                                                                                                  'fontWeight' : 'bold'},
                                                                                  style_cell = {'color' : 'white',
                                                                                                'backgroundColor' : 'rgb(50,50,50)',
                                                                                                'border' : '0px solid',
                                                                                                'font-family' : 'Open Sans Semi',
                                                                                                'font-size' : '0.8em',
                                                                                                'textAlign' : 'center',
                                                                                                'min-width' : '70px'},
                                                                                  style_as_list_view = True,
                                                                                  style_data_conditional = [{'if' : {'row_index' : 'odd'},
                                                                                                             'backgroundColor' : 'rgb(70,70,70)'}])
                                                                   ]
                                                          ),
                                                          html.Div('IP hlavička'),
                                                          html.Div(style = {'margin-bottom' : '10px'},
                                                                   children = [dash_table.DataTable(
                                                                                  id = 'IP_table',
                                                                                  fixed_rows = {'headers' : True,'data' : 0},
                                                                                  columns = [{'id' : 'version','name' : 'Verzia'},
                                                                                             {'id' : 'ttl','name' : 'TTL'},
                                                                                             {'id' : 'protocol','name' : 'Protokol'},
                                                                                             {'id' : 'srcIP','name' : 'IP adresa zdroja'},
                                                                                             {'id' : 'dstIP','name' : 'IP adresa cieľa'}],
                                                                                  style_table = {'border' : 'grey solid',
                                                                                                 'width' : 800,'height' : 60},
                                                                                  style_header = {'color' : 'lightgrey',
                                                                                                  'backgroundColor' : 'rgb(30,30,30)',
                                                                                                  'fontWeight' : 'bold'},
                                                                                  style_cell = {'color' : 'white',
                                                                                                'backgroundColor' : 'rgb(50,50,50)',
                                                                                                'border' : '0px solid',
                                                                                                'font-family' : 'Open Sans Semi',
                                                                                                'font-size' : '0.8em',
                                                                                                'textAlign' : 'center',
                                                                                                'min-width' : '70px'},
                                                                                  style_as_list_view = True,
                                                                                  style_data_conditional = [{'if' : {'row_index' : 'odd'},
                                                                                                             'backgroundColor' : 'rgb(70,70,70)'}])
                                                                   ]
                                                          ),
                                                          html.Div('TCP hlavička'),
                                                          html.Div(style = {'margin-bottom' : '20px'},
                                                                   children = [dash_table.DataTable(
                                                                                  id = 'TCP_table',
                                                                                  fixed_rows = {'headers' : True,'data' : 0},
                                                                                  columns = [{'id' : 'srcPORT','name' : 'Zdrojový PORT'},
                                                                                             {'id' : 'dstPORT','name' : 'Cieľový PORT'},
                                                                                             {'id' : 'seq','name' : 'SEQ'},
                                                                                             {'id' : 'ack','name' : 'ACK'}],
                                                                                  style_table = {'border' : 'grey solid',
                                                                                                 'width' : 800,'height' : 60},
                                                                                  style_header = {'color' : 'lightgrey',
                                                                                                  'backgroundColor' : 'rgb(30,30,30)',
                                                                                                  'fontWeight' : 'bold'},
                                                                                  style_cell = {'color' : 'white',
                                                                                                'backgroundColor' : 'rgb(50,50,50)',
                                                                                                'border' : '0px solid',
                                                                                                'font-family' : 'Open Sans Semi',
                                                                                                'font-size' : '0.8em',
                                                                                                'textAlign' : 'center',
                                                                                                'min-width' : '70px'},
                                                                                  style_as_list_view = True,
                                                                                  style_data_conditional = [{'if' : {'row_index' : 'odd'},
                                                                                                             'backgroundColor' : 'rgb(70,70,70)'}])
                                                                   ]
                                                          )
                                              ]
                                     )
                         ]
                )
    ]
)

# Callback funkcia pre tabulku 1 - "Obnovenie"
@app.callback(Output('table1','data'),[Input('obnov','n_clicks')])
def update_table1(n_c):
    # usporiadanie riadkov od najnovsieho paketu po najstarsi
    df = pkt_data.sort_values(["pktnum"],ascending = False)
    #print(df.to_dict('records'))
    return (df.to_dict('records'))

# Callback funkcia pre tabulku 1 - "Vyber bunky - riadku"
@app.callback(Output('selected_packet_text','children'),[Input('table1','active_cell')],[State('table1','data')])
def select_table1_row(active_cell,data):
    global sel_pkt

    if active_cell:
        row = active_cell['row']
        #col = active_cell['column_id']
        cell_data = data[row]['pktnum']
        sel_pkt = cell_data
        return html.P(f'Bol vybraný paket číslo: {sel_pkt}')
    
    return html.P('Nie je vybraný žiaden paket')

# Callback funkcia pre tabulky Ethernet, IP a TCP - "Detaily"
@app.callback([Output('Ethernet_table','data'),Output('IP_table','data'),Output('TCP_table','data')],[Input('detaily','n_clicks')])
def update_Ethernet_table(n_c):
    global sel_pkt

    # ak bol vyrany nejaky paket
    if sel_pkt != 0:
        # Ethernet hlavička
        Ethernet_df = Ethernet_data.loc[sel_pkt]
        Ethernet_df_list = [Ethernet_df.to_dict()]
        # IP hlavička - nemusi byt vyplnena
        IP_df_list = None
        if pkt_data.loc[sel_pkt].hasIP:
            IP_df = IP_data.loc[sel_pkt]
            IP_df_list = [IP_df.to_dict()]
        # TCP hlavička - nemusi byt vyplnena
        TCP_df_list = None
        if pkt_data.loc[sel_pkt].hasTCP:
            TCP_df = TCP_data.loc[sel_pkt]
            TCP_df_list = [TCP_df.to_dict()]
        return [Ethernet_df_list,IP_df_list,TCP_df_list]

    return [None,None,None]

# ukoncenie programu po stlaceni Ctrl+C
def CtrlC(signal,frame):
    print("Koniec...\n")
    sys.exit(0)

# callable object pre vlakno
def pktsniffer():
    print("Začiatok...")
    sniff(iface = None,prn = pkt_recv,filter = "ip",store = 0)

if __name__ == "__main__":
    signal.signal(signal.SIGINT,CtrlC) # nastavenie Signal handleru

    # sniffer v samostatnom vlakne
    t = threading.Thread(target = pktsniffer)
    t.daemon = True
    t.start()

    # spustenie Dash aplikacie
    app.run_server(debug = True)